
#include <fstream.h>
#include <time.h>
#include <Windows.h>

#include <cryptlib/cryptlib.h>
#include <cryptlib/des.h>
#include <cryptlib/misc.h>
#include <cryptlib/filters.h>
#include <cryptlib/modes.h>
#include <cryptlib/md5.h>

#include "WFCEncryptDecryptThread.h"
#include "WFC_Defines.h"

#ifdef DEBUG
#include "DEBUG.h"
#endif

#define NARGS 4


WFCEncryptDecryptThread::WFCEncryptDecryptThread()
{
	for(int i = 0; i < NARGS; i++)
		currentArgs[i] = NULL;

	encryptor = NULL;
	decryptor = NULL;

	lastErrorMessage = NULL;

	updateWin = NULL;
	updateWinPos.x = updateWinPos.y = 0;

	progress = 0;
}

WFCEncryptDecryptThread::~WFCEncryptDecryptThread()
{
	CleanUp();

	if(lastErrorMessage != NULL) {
		delete lastErrorMessage;
		lastErrorMessage = NULL;
	}
}

void WFCEncryptDecryptThread::CleanUp()
{
	for(int i = 0; i < NARGS; i++)
	{
		if(currentArgs[i] != NULL) {
			delete currentArgs[i];
			currentArgs[i] = NULL;
		}
	}

	if(encryptor != NULL) {
		delete encryptor;
		encryptor = NULL;
	}

	if(decryptor != NULL) {
		delete decryptor;
		decryptor = NULL;
	}

	if(infile.is_open())
	{
		infile.close();
	}

	if(outfile.is_open())
	{
		outfile.close();
	}
}

int WFCEncryptDecryptThread::LastError()
{
	return lastError;
}

void WFCEncryptDecryptThread::SetUpdateProgressBar(POINT* pos, HWND window)
{
	updateWin = window;
	updateWinPos = *pos;
}

void WFCEncryptDecryptThread::UpdateProgress()
{
	if(updateWin != NULL) {
		HDC hdevContext = GetDC(updateWin);
		RECT progbar_border;
		RECT progbar_inside;
		HBRUSH progbar_border_brush = (HBRUSH) GetStockObject(BLACK_BRUSH),
			progbar_filled_brush = CreateSolidBrush(RGB(0, 0, 0xFF)),
			progbar_clear_brush = (HBRUSH) GetStockObject(WHITE_BRUSH);

		SetRect(&progbar_border, updateWinPos.x, updateWinPos.y, updateWinPos.x + PROGBAR_WIDTH, updateWinPos.y + PROGBAR_HEIGHT);
		progbar_inside = progbar_border;
		progbar_inside.left +=1;
		progbar_inside.top +=1;
		progbar_inside.right -=1;
		progbar_inside.bottom -=1;

		//draw frame
		FrameRect(hdevContext, &progbar_border,  progbar_border_brush);

		float right = progress/100.0 * PROGBAR_WIDTH - 1;

		progbar_inside.right = progbar_inside.left + right + 1;
		FillRect(hdevContext, &progbar_inside, progbar_filled_brush);

		progbar_inside.left = progbar_inside.right;
		progbar_inside.right = progbar_border.right-1	;
		FillRect(hdevContext, &progbar_inside, progbar_clear_brush);

		//cleanup heap dynamic items
		DeleteObject(progbar_filled_brush);
	}
}

void WFCEncryptDecryptThread::init(bool doEncrypt, char* args[])
{
	CleanUp();

	progress = 0;

	int i;

	encrypt = doEncrypt;

	for(i = 0; i < NARGS; i++)
	{
		currentArgs[i] = new char [strlen(args[i]) + 1];
		memcpy(currentArgs[i], args[i], strlen(args[i]) + 1);
	}

	eof = dataWritten = 0;
	memset(checksum, 0, HASH_SIZE);
	memset(fchecksum, 0, HASH_SIZE);

	srand(time(NULL));

	memset(tkey, 0, KEYLEN);
	memcpy(tkey, currentArgs[1], strlen(currentArgs[1]) > KEYLEN ? KEYLEN : strlen(currentArgs[1]));

	if(doEncrypt) encryptor = new DES_EDE3_Encryption(tkey);
	else decryptor = new DES_EDE3_Decryption(tkey);

	md5.Restart();	//re-initialize MD5 instance

	infile.clear();
	outfile.clear();

	infile.open(currentArgs[2], ios::in|ios::binary|ios::nocreate);

	if(!infile.good()) {
		lastError = 1;
		SetLastErrorMessage("Failed to open input file...");
		CleanUp();
		return;
	}

	outfile.open(currentArgs[3], ios::out|ios::binary);

	if(!outfile.good()) {
		lastError = 1;
		SetLastErrorMessage("Failed to open output file...");
		CleanUp();
		return;
	}

	lastError = 0;
}


void WFCEncryptDecryptThread::DoEncryptDecrypt()
{
	int nread;

	//find file size
	infile.seekg(0, ios::end);
	eof = infile.tellg();		//wfc_eof is NUMBER of bytes in file...
	infile.seekg(0, ios::beg);

	if(encrypt)
	{
		//pad out to multiple of BUFSZ
		pad = BUFSZ - (eof % BUFSZ);
		if(pad == BUFSZ) pad = 0;
		outfile.write(&pad, sizeof(pad));


		//fill IV with random stuff
		MemSetRand(prevblock, BUFSZ);
		outfile.write(prevblock, BUFSZ);


		while(dataWritten < eof)
		{
			MemSetRand(eblock, BUFSZ);

			infile.read(curblock, BUFSZ);
			nread = infile.gcount();


			//update md5 sum with this block of data
			md5.Update(curblock, nread);


			//encrypt block
			XORBlock(prevblock, curblock, BUFSZ);	//cipher-block-chaining
			encryptor->ProcessBlock(curblock, eblock);
			memcpy(prevblock, eblock, BUFSZ);


			if(nread < BUFSZ) {	//end of file include padding
				outfile.write(eblock, nread + pad);
			}
			else {
				outfile.write(eblock, nread);
			}


			dataWritten += nread;


			//update progress bar
			int nprogress = (float) dataWritten / (float) eof * 100;
			if(nprogress != progress) {
				progress = nprogress;
				UpdateProgress();
			}

		}

		//at the end encrypt and write the checksum
		md5.TruncatedFinal(checksum, sizeof(checksum));	//get checksum

		//encrypt/write out blocks of checksum
		for(int i = 0; i < sizeof(checksum); i += BUFSZ)
		{
			memcpy(curblock, checksum + i, BUFSZ);

			XORBlock(prevblock, curblock, BUFSZ);
			encryptor->ProcessBlock(curblock, eblock);

			memcpy(prevblock, eblock, BUFSZ);

			outfile.write(eblock, BUFSZ);
		}
	}

	
	//decryption
	if(!encrypt)
	{
		//read in padding size
		infile.read(&pad, sizeof(pad));
		
		unsigned long otherBytes = sizeof(pad) + pad + IV_SIZE + sizeof(fchecksum);	//WFC data
		unsigned long dataBytes = eof - otherBytes;	//actual data to be written out

		//check for corrupted file
		if(((dataBytes+pad) % BUFSZ) != 0) {
			lastError = 1;
			SetLastErrorMessage("Invalid padding amount, file corrupt...");

			outfile.close();
			DeleteFile(currentArgs[3]);

			CleanUp();
			return;
		}

		//read in Initialization Vector
		infile.read(prevblock, IV_SIZE);


		//write original decrypted data out to file
		while(dataBytes > 0)
		{
			memset(curblock, 0, BUFSZ);

			infile.read(eblock, BUFSZ);
			nread = infile.gcount();

			
			//decrypt data
			decryptor->ProcessBlock(eblock, curblock);
			XORBlock(prevblock, curblock, BUFSZ);	//cipher block chained
			memcpy(prevblock, eblock, BUFSZ);


			int wbytes = (dataBytes < BUFSZ)? nread-pad : nread;	//num of bytes to write

			outfile.write(curblock, wbytes);


			//update MD5 sum with this data
			md5.Update(curblock, wbytes);


			dataBytes -= wbytes;

			//update progress bar
			int nprogress = ( (float) ((eof-otherBytes) - dataBytes) / (float) (eof-otherBytes)) * 100;
			if(nprogress != progress) {
				progress = nprogress;
				UpdateProgress();
			}
		}


		//now read checksum
		infile.read((char*) &fchecksum, sizeof(fchecksum));

		//decrypt MD5 checksum
		for(int i = 0; i < sizeof(fchecksum); i += BUFSZ)
		{
			memcpy(eblock, fchecksum + (i), BUFSZ);

			decryptor->ProcessBlock(eblock, curblock);
			XORBlock(prevblock, curblock, BUFSZ);

			memcpy(fchecksum + i, curblock, BUFSZ);

			memcpy(prevblock, eblock, BUFSZ);
		}

		md5.TruncatedFinal(checksum, sizeof(checksum));

		//compare decrypted file's checksum with our checksum
		if(memcmp(checksum, fchecksum, sizeof(checksum)) != 0)
		{
			lastError = 1;
			SetLastErrorMessage("Checksum mismatch!");
			outfile.close();
			DeleteFile(currentArgs[3]);

			progress = 0;
			UpdateProgress();

			CleanUp();
			return;
		}
	}

	//successfully encrypted/decrypted file
	infile.close();
	DeleteFile(currentArgs[2]);	//delete original

	CleanUp();
}

void WFCEncryptDecryptThread::SetLastErrorMessage(char* msg)
{
	if(lastErrorMessage != NULL)
		delete lastErrorMessage;

	lastErrorMessage = new char [strlen(msg)];
	strcpy(lastErrorMessage, msg);
}

void WFCEncryptDecryptThread::GetLastErrorMessage(char* dest, int maxlen)
{
	if(lastErrorMessage != NULL)
		if(strlen(dest) < maxlen)
			strcpy(dest, lastErrorMessage);
}

int WFCEncryptDecryptThread::GetProgress()
{
	return progress;
}